/*
level_names example: ['level_0_title', 'level_1_title', 'level_2_title', ...]
measure_1 example: 'budget_usd'. Will be used for sizing the rects in the tree.
measure_2 example: 'actual_usd'. Can be used to compute a actual/budget ratio.
*/ 

function buildHierarchy(flat_data, level_names, measure, color_measure=null) {
  
  var nest = d3.nest()
  
	//SET A GRANDPARENT 
  nest = nest.key(function(d) {return 'All';});
  
  
  //SET ALL LEVELS SPECIFIED BY LEVEL_NAMES
  level_names.forEach(function(level) {
    nest = nest.key(function(d) { return d[level]});
  });
    
  
  //FILL NESTED STRUCTURE WITH FLAT DATA
  const nested_data = nest.entries(flat_data)[0];
	
  
  //TRANSFORM NESTED DATA INTO HIERARCHICHAL STRUCTURE
  var root = d3.hierarchy(nested_data, (d) => d.values)
    // Computes a 'value' property for all nodes.
    .sum(function(d) {return d[measure]}) 
    .sort(function(a, b) {return d3.descending(a.value, b.value) });
  


  
  
  //CALCULATE PERCENTAGE OF TOTAL
  root.each(function(node){
    
    if (node.depth >1) {
      node["Remaining_Days"] = parseInt( node.leaves()[0].data.Remaining_Days )
    }; 
    
    if (node.depth >= 1 && node.depth <=2) {
      node['Percent_Total_Level'] = node.value/node.parent.value
    };
      
    if (color_measure) { // **IF ARG COLOR METRIC EXISTS
      node['color_metric'] = node[color_measure];
    };
    
  });
  
  
  
  //IN CASE AN ITEM ONLY HAS ONE LOT, DELETE ITS CHILDREN IN THE LAST LEVEL
  //Prune leaves in these cases
  var replace_array
  root.each(function(node) {
    //if the node has only one child, move child´s data up one level and delete child
    if (node.depth == 1 && node.children.length==1){       
      replace_array = node.leaves()[0].data
      node.children[0].data = replace_array;
      delete node.children[0].children;
    }   
  });

	console.log(root)
  return root;
}
  

function drawChart(hierData,level_names) {
  // Clear any existing svg elements
  //d3.selectAll('svg').remove();
  

  var margin = {top: 20, right: 0, bottom: 0, left: 0},
      width = window.innerWidth,
      height = window.innerHeight - margin.top - margin.bottom,
      formatNumber = d3.format(",d"),
      padding = 0,
      transitioning;

  var x = d3.scaleLinear()
      .domain([0, width])
      .range([0, width]);

  var y = d3.scaleLinear()
      .domain([0, height])
      .range([0, height]);

  // A function that computes x0, x1, y0, y1 for nodes in a given hierarchy.
  const treemap = d3.treemap().size([width, height])
     .paddingOuter(padding);

  
  
  var svg = d3.select("body").append("svg")
  .attr("width", width + margin.left + margin.right)
      .attr("height", height + margin.bottom + margin.top)
      .style("margin-left", -margin.left + "px")
      .style("margin.right", -margin.right + "px")
    .append("g")
      .attr("transform", "translate(" + margin.left + "," + margin.top + ")")
      .style("shape-rendering", "crispEdges");
  

	//Creating and setting grandparent rect
  var grandparent = svg.append("g")
      .attr("class", "grandparent");

  grandparent.append("rect")
      .attr("x", padding * 2)
      .attr("y", -margin.top)
      .attr("width", width - padding * 4)
      .attr("height", margin.top);

  grandparent.append("text")
      .attr("x", 6)
      .attr("y", 6 - margin.top)
      .attr("dy", ".75em")
  		.attr("fill", "white");
  


	//Executing functions before defining them!
  
  function display(d) {
    treemap(hierData);
  	
    //Defining function that returns the domain for scaleLinear()
    function colorDomain() {
      var root = hierData;
      color_values = [];

      //Just useful for debug
      if (!root.children) {
        color_values = [root.color_metric];
        return;
      }

      //Get an array with the color metric of every leaf element
      root.children.forEach(child => {
        if (child.children ) {
          child.children.forEach(grandchild => {
            color_values.push(grandchild.color_metric);
          });
        } else {
          color_values.push(child.color_metric);
        }
      });

      return [d3.min(color_values), d3.mean(color_values), d3.max(color_values)];    
  	}
    
    
    //TO DO: MOVE OUT OF FUNCTION AND SET AS ARGUMENT

    
    var color_range_object =
    {LEVEL_1:["#4485cf","#84abe7","#bad3ff"],
     LEVEL_2:["#f2e1ff","#cfb6ea","#a086c9"],
     LEVEL_3:["#ff9bc6","#ffbada","#ffd7ec"],
     LEVEL_4:["#ff6973","#ffa3a3","#ffd7d6"],
     OK: ["#f5faff", "#dcf0ff", "#bfe7ff"]}
    
    
    //loop through each level´s color range to:
    // - create an instance of scaleLinear for that level
    // - store a color accessor function that uses the defined scale
    var level_colors_array = Object.values(color_range_object)
    var scale_instances_array = []
    var color_accessor_array = {}
    var color_accessor_key

    level_colors_array.forEach(function(d,i){
      scale_instances_array[i] = d3.scaleLinear()
      .range(level_colors_array[i])
      .domain(colorDomain());
  
      color_accessor_key = Object.keys(color_range_object)[i]
      color_accessor_array[color_accessor_key] = function(d){return scale_instances_array[i](d.color_metric)}
    })
  
  
    //TO DO: MAKE TOOLTIP USE THIS
    var color = d3.scaleLinear()
      .range(level_colors_array[1])
      .domain(colorDomain())
    
    grandparent
        // Attach zoom function to grandparent -- works after 1st zoom, when node has 'parent'
        .datum(d.parent) //enables return to parent when clicked
        .on("click", zoom)
      .select("text")
        // Name for grandparent is the hierarchical path
        .text(name(d));
    grandparent.select('rect')
        .style("fill", "navy");
    
    // g1 = main graphical element to hold the tree, inserted after the top g.
    // g1 is "the grandparent", even though the top g holds the text about it.
    var g1 = svg.insert("g", ".grandparent")
        .datum(d) // Gets the whole hierarchical structure for svg
        .attr("class", "depth");
    
    // Within g1, create g's for the "parents" (children of grandparent)
    var g = g1.selectAll("g")
        .data(d.children)
      .enter().append("g");
    
    // Attach zoom function to parents who have children.
    g.filter(function(d) { return d.children; })
        .attr("hasChildren", true)
        .on("click", zoom);
    
    // Within each parent g, add g's to hold rect + text for each child.
    var children = g.selectAll(".child")
        .data(function(d) {return d.children || [d]; })
      .enter().append("g");
    
   
    //looping through number of levels to use the d3 filter function
    //enables a color per level
    var root = hierData
    var number_levels = root.children.length

    
    for (i=0;i<number_levels;i++){
      
      //if leaf of rect belongs to a given alert level, use 
      children.filter(
        function(d){
          console.log(level_names)
          return d.leaves()[0].data.Alert_Level == root.data.values[i].key;
        })
      .append("rect")
      .attr("class", "child")
      .call(rect)
      .style("fill", color_accessor_array[root.data.values[i].key])
      .on("mousemove", d => mousemove(d))
      .on("mouseout", mouseout);
    }
      


    
    

		//Add text to each child rect
    children.append("text")
        .attr("class", "ctext")
        .text(function(d) {return d.data.key;})
        .call(childrenText); //the function call sets position of text in bottom right side of rect
    
    var set_text = function (d) {} // DEFINE FUN THAT RETURNS A STRING WITH LEAF DATA VALUES IF PASSED OBJECT HAS NO CHILDREN
    
    
    
    var get_node_text = function(d) {
      if(!d.children){
				return
      }
      return (formatNumber(d.value))
    };
    
    var get_leaf_text = function(d) {
      //If node is a leaf, then return special text
      if(!d.children){
      	var leaf_string_array = []  
       	Object.entries(d.data).forEach(function(e) {leaf_string_array.push(e[0] + ": " +e[1])})
        return (d.data.Label_1)
      } 
    }
    
    

    // Even though all children are rects, parents are also presented as rects.
    g.append("rect")
        .attr("class", "parent")
        .call(rect); //called function sets dimensions and position of rect

    var t = g.append("text")
        .attr("class", "ptext")
        .attr("dy", ".75em")
    
    t.append("tspan")
        .text(function(d) { return d.data.key; });
    
    t.append("tspan")
        .attr("dy", "1.0em")
        .text(get_node_text)
    
    
    
    //If current selection is a leaf, then get an array of strings with data, and add them as svg text
    var leaf_data = d.leaves()[0].data  
    var leaf_array = []
    if(d.children.length == 1){
      t.style("font-size", "18px")
      
      var leaf_data = d.leaves()[0].data  
      var leaf_array = []
      Object.entries(leaf_data).forEach(
        function(e) {
          leaf_array.push(e)
        }
      )
  	//loop through array of data fields to print them on screen
    leaf_array.forEach(e=> {
       t.append("tspan")
        .attr("dy", "1.5em")
        .text(e[0] + " : " + e[1])
      	.attr("fill", "dark_grey")
     }) 
    }
    
    t.call(parentText);
    


    // Draw/move child nodes on top of parent nodes.
    svg.selectAll(".depth").sort(function(a, b) {return a.depth - b.depth;});

    // FIXME: Move to private function 
    function zoom(d) {
      if (transitioning || !d) return;
      transitioning = true;
	
      var g2 = display(d);
      // At this point in time, g1 with its rects are still in place.
      // The parent that has just been promoted to grandparent now has 2 levels
      // of rects visualized, but still positioned exactly where it was.

      var t1 = g1.transition().duration(750);
      var t2 = g2.transition().duration(750);

      // Enable anti-aliasing during the transition.
      svg.style("shape-rendering", null);

      // Update the domain only after entering new elements (within display()).
      x.domain([d.x0, d.x1]);
      y.domain([d.y0, d.y1]);

      // This moves all rects into the right position & width/height.
      t1.selectAll("rect").call(rect);
      t2.selectAll("rect").call(rect);

      // Fade-in entering text.
      g2.selectAll("text").style("fill-opacity", 0);

      // Transition texts to the new view.
      t1.selectAll(".ptext").call(parentText).style("fill-opacity", 0);
      t1.selectAll(".ctext").call(childrenText).style("fill-opacity", 0);
      t2.selectAll(".ptext").call(parentText).style("fill-opacity", 1);
      t2.selectAll(".ctext").call(childrenText).style("fill-opacity", 1);

      // Remove the old node when the transition is finished.
      t1.remove().on("end", function() {
        svg.style("shape-rendering", "crispEdges");
        transitioning = false;
      });
    }// End of zoom()
    

    
    function parentText(text) {
      var min_text_length = 6
      var min_rect_height = 50
      text.selectAll("tspan")
      		.attr("font-size", "12px")
          .attr("x", function(d) { return x(d.x0) + 6; });
      text.attr("x", function(d) { return x(d.x0) + 6; })
          .attr("y", function(d) { return y(d.y0) + 6; })
					.style("opacity", function(d) {return (this.getComputedTextLength() < (x(d.x1) - x(d.x0) - min_text_length)  
                                         && (d.y1- d.y0 > min_rect_height)) || d.depth > 2 ? 1 : 0; })
    }
      	
      		
    function childrenText(text) {
      var min_text_length = 6
      var min_rect_height = 25
      text.attr("x", function(d) { return x(d.x1) - this.getComputedTextLength() - 6; })
          .attr("y", function(d) { return y(d.y1) - 6; })
      	  .style("opacity", function(d) {return (this.getComputedTextLength() < x(d.x1) - x(d.x0) - min_text_length  
                                         && (d.y1- d.y0 > min_rect_height)) ? 1 : 0; })
    }



    function rect(rect) {
      rect.attr("x", function(d) { return x(d.x0); })
          .attr("y", function(d) { return y(d.y0); })
          .attr("width", function(d) { console.log(x(d.x1)); return x(d.x1) - x(d.x0); })
          .attr("height", function(d) { return y(d.y1) - y(d.y0); });
    }
    

    function name(d) {
      return d.parent
          ? name(d.parent) + " > " + d.data.key
          : d.data.key;
  	}

    return g;
  }//End of display()
	display(hierData);

	
    //Creating tooltip
    //Because r2d3 only shows svg, tooltip has to be a special div appended to svg
   var foreign = svg.append("foreignObject")
      .attr("class", "svg-tooltip")
      .attr("height", "100")
      .attr("width", "300")
   		.style("display", "none");

   var tooltip_foreign = foreign.append("xhtml:div")
      .attr("class", "tooltip")
      .style("display","table-cell")	
      .style("text-align", "center")
      .style("vertical-align", "middle")
   		.style("line-height", "0.2em")
   		.style("text-align", "left");  

   foreign.raise()

  
  //Creating tooltip divs and ps  	
  const tooltip_heading = tooltip_foreign.append("p")
  	.attr("id", "tooltip-title")
  	.style("font-weight", "bold");
  
  for (i=1; i<5; i++){
  var tooltip_text = tooltip_foreign.append("p");
  tooltip_text.append("span")   
    .attr("id", "tooltip-measure-" + i + "-key")
    .style("font-weight", "bold");
  
  tooltip_text.append("span")   
    .attr("id", "tooltip-measure-" + i + "-value");  
  
  };

 
  
  function mousemove(d) {
    
    var tooltip_height = tooltip_foreign.node().getBoundingClientRect().height
    var tooltip_width = tooltip_foreign.node().getBoundingClientRect().width
    var mouse_x = d3.mouse(d3.event.target)[0]
		var mouse_y = d3.mouse(d3.event.target)[1]

    
    //functions that return tooltip position relative to mouse coordinates and distance from screen
    var xPosition = function() {
      if ( mouse_x  > width/2) {
        return d3.event.pageX - 200;     
      } else {
      	return d3.event.pageX + 100;
      }
    }
    var yPosition =function() { 
      if (mouse_y > height/2){
      	return d3.event.pageY - 150
      } else {
      return d3.event.pageY + 75;}};
    
 
    
    //Setting position of tooltip
    foreign
    	.transition()
    	.duration(100)
      //.style("background-color", color_accessor_array[level_index])
      .attr("x", xPosition())
      .attr("y", yPosition());
   
    //Setting text of tooltip
    tooltip_heading
      .text(d.leaves()[0].data.Item_No);
    
    d3.select("#tooltip-measure-1-value")
    	.text(d.leaves()[0].data.Label_1)
    	.style("font-weight", "bold");
    
    d3.select("#tooltip-measure-2-key")
    	.text("Remaining_Quantity: ")
    	.style("font-weight", "bold");
    
    d3.select("#tooltip-measure-2-value")
    	.text(d3.format(",.0f")
            (d.leaves()[0].data.Lot_Remaining_Quantity));
    
    d3.select("#tooltip-measure-3-key")
    	.text("Remaining_Days: ")
    	.style("font-weight", "bold");
    
    d3.select("#tooltip-measure-3-value")
    	.text(d3.format(",.0f")
            (d.leaves()[0].data.Remaining_Days));
    
    d3.select("#tooltip-measure-4-key")
      .text("% of Level: ")
      .style("font-weight", "bold");
    
    d3.select("#tooltip-measure-4-value")
    	.text(d3.format(",.1%")(d.Percent_Total_Level));
    
    //Get rid of level metric in tooltip beyond first node  
    if (d.depth>2){
      d3.select("#tooltip-measure-4-key")
        .text("")
      d3.select("#tooltip-measure-4-value")
        .text("");
      }
    
    foreign.style("display", "block") 
    tooltip_foreign   
      .transition().duration(50).style("opacity", "0.85");
    
  };

  function mouseout() {
    tooltip_foreign
      .transition()
      .duration(200)
      .style("opacity","0");
    foreign.style("display", "none")  ;
  };
} //End of drawChart()



  
d3.csv('stock.csv', function(error, data) {
   
  var hier = [
    'Alert_Level',
    'Item_No',
    'Remaining_Days'
  ];
  
  var hierarchical_data = 
    buildHierarchy(
      data, 
      hier, 
      'Lot_Remaining_Quantity', 
      'Remaining_Days' // color_measure 
    );

  drawChart(
    hierarchical_data, hier
  );
});